package SliMP3::Info;

# SliMP3 Server Copyright (C) 2001 Sean Adams, Slim Devices Inc.
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License,
# version 2.

use strict;


use File::Spec::Functions qw(:ALL);
use MLDBM qw(SDBM_File Storable);
use Fcntl;

use MP3::Info;
use SliMP3::Ogg;
use SliMP3::Misc;
use SliMP3::OSDetect;
use SliMP3::Strings qw(string);
#
# constants
#

# the items in the infocache that we actually use
# NOTE: THE ORDER MATTERS HERE FOR THE PERSISTANT DB
# IF YOU ADD SOMETHING, PUT IT AT THE END
my @infoCacheItems = (
	'CT',	 # content type
	'TITLE', # title
	'LIST',	 # list items (array)
	'AGE',   # list age
	'GENRE', # genre
	'TRACKNUM', # tracknumber as an int
	'FS',	 # file size
	'SIZE',	 # audio data size in bytes
	'OFFSET', # offset to start of song
	'ARTIST', # artist
	'ALBUM',  # album name
	'COMMENT',	# ID3 comment
	'YEAR',		# year
	'SECS', 	# total seconds
	'VBR_SCALE', # vbr/cbr
	'BITRATE', # bitrate
	'TAGVERSION', # ID3 tag version
	'COMPOSER', # composer
	'TAGSIZE', # tagsize
	'DISC', # album number in a set 
	'DISCC', # number of albums in a set
);
my %infoCacheItemsIndex;

# the types we know about.
my @types = ( 'mp2', 'mp3', 'ogg', 'dir', 'm3u', 'pls', 'lnk', 'cue', 'itu');

# please add mime types to this list as you encounter them in the wild.
my %mimeTypes = (
	'audio/x-mpegurl' => 'm3u',
	'audio/mpegurl'   => 'm3u',
	'audio/mpeg'	  => 'mp3',
	'audio/x-scpls'   => 'pls',
	'audio/x-ogg'	  => 'ogg'
);

#
# global caches
#

# a hierarchical cache of genre->artist->album->tracknum based on ID3 information
my %genreCache = ();

# a cache of the titles used for uniquely identifing and sorting items 
my %caseCache = ();

# the main cache of ID3 and other metadata
my %infoCache = ();
my %infoCacheDB = (); # slow, persistent cache structure

##################################################################################
# these routines deal with the caches directly
##################################################################################
sub init {

	my $dbname;

	my $i = 0;
	foreach my $tag (@infoCacheItems) {
		$infoCacheItemsIndex{$tag} = $i;
		$i++;
	}
	
	if (SliMP3::Prefs::get('usetagdatabase')) {
		$dbname = catdir(SliMP3::Prefs::preferencesPath(), "mp3tag.db");

		# TODO: MacOS X should really store this in a visible, findable place.
		if (SliMP3::OSDetect::OS() eq 'unix') {
			# try the default one (in the home directory) should be read...
			$dbname = catdir(SliMP3::Prefs::preferencesPath(), '.slimp3info.db');
		} elsif (SliMP3::OSDetect::OS() eq 'win')  {
			$dbname = catdir(SliMP3::Prefs::preferencesPath(), 'SLIMP3INFO.DB');
		} else {
			die "Unknown OS type, don't know what style name to save.";
		}

		$::d_info && SliMP3::Misc::msg("ID3 tag database support is ON, saving into: $dbname\n");

		tie (%infoCacheDB, 'MLDBM', $dbname, O_CREAT|O_RDWR, 0666)
			or warn "Error opening tag database $dbname: $!";

		foreach my $file (keys %infoCacheDB) {
			if (isSong($file)) {
				my $cacheEntryArray =  $infoCacheDB{$file};
				my $cacheEntryHash;

				my $i = 0;
				foreach my $key (@infoCacheItems) {
					$cacheEntryHash->{$key} = $cacheEntryArray->[$i];
					$i++;
				}

				updateGenreCache($file, $cacheEntryHash);
			}
		}
		$::d_info && SliMP3::Misc::msg("done loading genre cache from DB\n");
	}
	
	# use all the genres we know about...
	MP3::Info::use_winamp_genres();
	
	#turn on unicode support
#	if (!MP3::Info::use_mp3_utf8(1)) {	
#		$::d_info && SliMP3::Misc::msg("Couldn't turn on unicode support.\n");
#	};
}

sub stopCache {
	untie (%infoCacheDB);
}

sub clearCache {
	my $item = shift;
	if ($item) {
		$infoCache{$item} = undef;
		$::d_info && SliMP3::Misc::msg("cleared $item from cache\n");
	} else {
		%infoCache = ();
		# a hierarchical cache of genre->artist->album->song based on ID3 information
		%genreCache = ();
		%caseCache = ();
	}
}

sub isCached {
	my $url = shift;
	return exists $infoCache{$url};
}

sub cacheItem {
	my $url = shift;
	my $item = shift;
	my $cacheEntryArray;
	
	if (!defined($url)) {
		SliMP3::Misc::msg("Null cache item!\n"); 
		SliMP3::Misc::bt();
		return undef;
	}
	
	if (exists $infoCache{$url}) {
		$cacheEntryArray = $infoCache{$url};
		my $index = $infoCacheItemsIndex{$item};
		if (exists $cacheEntryArray->[$index]) {
			return $cacheEntryArray->[$index];
		} else {
		}	return undef;
	}

	# cache miss from memory, check the disk
	if (SliMP3::Prefs::get('usetagdatabase') &&
			!defined $cacheEntryArray &&
			exists $infoCacheDB{$url}) {
		$cacheEntryArray = $infoCacheDB{$url};
		$infoCache{$url} = $cacheEntryArray;
		my $index = $infoCacheItemsIndex{$item};
		if (defined $cacheEntryArray && exists $cacheEntryArray->[$index]) {
			return $cacheEntryArray->[$index];
		}
	}
	
	return undef;
}

sub cacheEntry {
	my $url = shift;
	my $cacheEntryHash;
	my $cacheEntryArray;

	if ($::d_info && !defined($url)) {die;}
	
	if ( exists $infoCache{$url}) {
		$cacheEntryArray = $infoCache{$url};
	};

	# cache miss from memory, check the disk
	if (SliMP3::Prefs::get('usetagdatabase') &&
			!defined $cacheEntryArray &&
			exists $infoCacheDB{$url}) {
		$cacheEntryArray = $infoCacheDB{$url};
		$infoCache{$url} = $cacheEntryArray;
	}
	
	my $i = 0;
	foreach my $key (@infoCacheItems) {
		if ($cacheEntryArray->[$i]) {
			$cacheEntryHash->{$key} = $cacheEntryArray->[$i];
		}
		$i++;
	}

	return $cacheEntryHash;
}

sub updateCacheEntry {
	my $url = shift;
	my $cacheEntryHash = shift;
	my $cacheEntryArray;
	
	if ($::d_info && !defined($url)) { SliMP3::Misc::msg(%{$cacheEntryHash}); SliMP3::Misc::bt();  die; }
	if ($::d_info && !defined($cacheEntryHash)) {SliMP3::Misc::bt(); die; }

	if (!defined($url)) { return; }
	if (!defined($cacheEntryHash)) { return; }

	#$::d_info && SliMP3::Misc::bt();
	
	if (SliMP3::Prefs::get('useinfocache')) {
		# if we've already got this in the cache, then merge the new over the old
		if (defined($infoCache{$url})) {
			my %merged = (%{cacheEntry($url)}, %{$cacheEntryHash});
			$cacheEntryHash = \%merged;
			$::d_info && SliMP3::Misc::msg("merging $url\n");
		}

		my $i = 0;
		foreach my $key (@infoCacheItems) {
			$cacheEntryArray->[$i] = $cacheEntryHash->{$key};
			$i++;
			$::d_info && $cacheEntryHash->{$key} && 
				SliMP3::Misc::msg("updating $url with " . $cacheEntryHash->{$key} . " for $key\n");
		}

		$infoCache{$url} = $cacheEntryArray;
		if (SliMP3::Prefs::get('usetagdatabase')) {
			$infoCacheDB{$url} = $cacheEntryArray;
		}
	}
}
##################################################################################

sub setContentType {
	my $url = shift;
	my $type = shift;

	my $cacheEntry = cacheEntry($url);

	$type = lc($type);
	if ($mimeTypes{$type}) {
		$type = $mimeTypes{$type};
	}

	$cacheEntry->{'CT'} = $type;

	updateCacheEntry($url, $cacheEntry);

	$::d_info && SliMP3::Misc::msg("Content type for $url is cached as $type\n");
}

my $ncElemstring = "VOLUME|PATH|FILE|EXT|DURATION|LONGDATE|SHORTDATE|CURRTIME"; #non-cached elements
my $ncElems = qr/$ncElemstring/;

my $elemstring = (join '|',@infoCacheItems,$ncElemstring);
#		. "|VOLUME|PATH|FILE|EXT" #file data (not in infoCache)
#		. "|DURATION" # SECS expressed as mm:ss (not in infoCache)
#		. "|LONGDATE|SHORTDATE|CURRTIME" #current date/time (not in infoCache)
my $elems = qr/$elemstring/;


#TODO Add elements for size dependent items (volume bar, progress bar)
#my $sdElemstring = "VOLBAR|PROGBAR"; #size dependent elements (also not cached)
#my $sdElems = qr/$sdElemstring/;

sub elemLookup {
	my $element = shift;
	#my $file = shift;
	my $infoHashref = shift;
	my $value;

	$value = $infoHashref->{$element};
	if ($element eq 'TRACKNUM') {
		#treat tracknumbers of 0 as undef
		undef $value unless $value;
	}

	return $value;
}

# used by infoFormat to add items not in infoCache to hash of info
sub addToinfoHash {
	my $infoHashref = shift;
	my $file = shift;
	my $str = shift;
	
	if ($str =~ /VOLUME|PATH|FILE|EXT/) {
		my ($volume, $path, $filename) = splitpath($file);
		$filename =~ s/\.([^\.]*?)$//;
		my $ext = $1;
		$infoHashref->{'VOLUME'} = $volume;
		$infoHashref->{'PATH'} = $path;
		$infoHashref->{'FILE'} = $filename;
		$infoHashref->{'EXT'} = $ext;
	}
	if ($str =~ /LONGDATE|SHORTDATE|CURRTIME/) {
		$infoHashref->{'LONGDATE'} = SliMP3::Misc::longDateF();
		$infoHashref->{'SHORTDATE'} = SliMP3::Misc::shortDateF();
		$infoHashref->{'CURRTIME'} = SliMP3::Misc::timeF();
	}
	if ($str =~ /DURATION/ && defined($infoHashref->{'SECS'})) {
		$infoHashref->{'DURATION'} = int($infoHashref->{'SECS'}/60) . ":" 
			. ((($infoHashref->{'SECS'}%60) < 10) ? ("0" . $infoHashref->{'SECS'}%60) : $infoHashref->{'SECS'}%60)
	}
}

#formats information about a file using a provided format string
sub infoFormat {
	no warnings; # this is to allow using null values with string concatenation, it only effects this procedure
	my $file = shift; # item whose information will be formatted
	my $str = shift; # format string to use
	my $safestr = shift; # format string to use in the event that after filling the first string, there is nothing left
	my $pos = 0; # keeps track of position within the format string

	my %infoHash = %{infoHash($file)}; # hash of data elements not cached in the main repository
	if (!defined($str)) { #use a safe format string if none specified
		$str = 'TITLE';
	}
	if ($str =~ $ncElems) {
		addToinfoHash(\%infoHash,$file,$str);
	}

	#here is a breakdown of the following regex:
	#\G -> start at the current pos() for the string
	#(.*?)? -> match 0 or 1 instances of any string 0 or more characters in length (non-greedy),capture as $1
	#(?:=>(.*?)=>)? -> 0 or 1 instances of any string in a =>=> frame, capture as $2 (excluding =>=>)
	#(?:=>(.*?)=>)? -> same, captured as $3
	#($elems) -> match one of the precompiled list of allowed data elements, capture as $4
	#(?:<=(.*?)<=)? -> 0 or 1 instances of any string in a <=<= frame, capture as $5 (excluding <=<=)
	#(.*?)? -> 0 or 1 instances of any string, capture as $6
	#(?:<#(.*?)#>)? -> 0 or 1 instances of any string in a <##> frame, capture as $7 (excluding <##>)
	#($elems|(?:=>.*?=>)) -> either another data element or a =>=> frame in front of another data element, as $8 (includes =>=>)
	while ($str =~ s{\G(.*?)?(?:=>(.*?)=>)?(?:=>(.*?)=>)?($elems)(?:<=(.*?)<=)?(.*?)?(?:<#(.*?)#>)?($elems|(?:=>.*?=>))}
				{
					my $out = ''; #replacement string for this substitution
					#look up the value corresponding to the first data element
					my $value = elemLookup($4,\%infoHash);
					#if another data element comes next replace <##> frames with =>=> frames
					#otherwise leave off the frames
					my $frame = defined($8) ? "=>" : "";
					if (defined($value)) { #the data element had a value, so include all the separators
						#pos() is set to the first character of the match by the s/// function
						#so adjust it to be where we want the next s/// to start
						#$pos is used to hold the value of pos() that we want, because pos()
						#gets reset to 0 since we aren't using /g.  We aren't using /g because we need
						#do do some backtracking and that is not allowed all in one go.
						#we want the next replace to start at the beginning of either
						#the next data element or the =>=> frame preceding it
						$pos = pos($str) + length($1) + length($2) + length($3) + length($value);
						if (!(defined($5) || defined($7))) {
							#neither a <=<= or a <##> frame is present so treat a bare
							#separator like one framed with <##>
							$out = $1 . $2 . $3 . $value . $frame . $6 . $frame . $8;
						} else {
							#either <=<= or <##> was present, so always write out a bare separator
							#if no <##> was present, don't add a =>=> frame
							$out = $1 . $2 . $3 . $value . $5 . $6 . (defined($7) ? ($frame . $7 . $frame) : "") . $8;
							#we want the next replace to start at the beginning of either
							#the next data element or the =>=> frame preceding it
							$pos += length($5) + length($6);
						}
					} else { #the data element did not have a value, so collapse the string
						#initialize $pos
						$pos = pos($str);
						if (defined($2) || defined($3)) {
							#a =>=> frame exists so always write a preceding bare separator
							#which should only happen in the first iteration if ever
							$out = $1;
							$pos += length($1);
							if (defined($6)) {
								#the bare separator is always used if a <=<= or a <##> frame was present
								#otherwise since there was a =>=> frame preceding the missing element convert
								#a bare separator to a =>=>
								$out .= (defined($5) || defined($7)) ? $6 : $frame . $6 . $frame;
								$pos += (defined($5) || defined($7)) ? length($6) : 0;
							}
							if (defined($7)) {
								#since there was a =>=> frame preceding the missing element convert
								#the <##> separator to a =>=>
								$out .= $frame . $7 . $frame;
							}
						} else {
							#treat a non-zero length bare separator as a data element for the purpose of determining
							#whether to convert a bare separator or a <##> to a =>=>
							$out = "";
							if (defined($6)) {
								$out .= (defined($5) || defined($7)) ? $6 : (length($1) ? ($frame . $6 . $frame) : "");
								$pos += (defined($5) || defined($7)) ? length($6) : 0;
							}
							if (defined($7)) {
								$out .= (length($1) || length($6)) ? ($frame . $7 . $frame) : "";
							}
						}
						$out .= $8;
					}
					$out;
				}e) {
		# since we aren't using s///g we need to reset the string position after each pass
		pos($str) = $pos;
	}
	# reset the string position which the failed match set to 0
	pos($str) = $pos;
	#same regex as above, but the last element is the end of the string
	$str =~ s{\G(.*?)?(?:=>(.*?)=>)?(?:=>(.*?)=>)?($elems)(?:<=(.*?)<=)?(.*?)?(?:<#(.*?)#>)?$}
				{
					my $out = '';
					my $value = elemLookup($4,\%infoHash);
					if (defined($value)) {
						#fill with all the separators
						#no need to do the <##> conversion since this is the end of the string
						$out = $1 . $2 . $3 . $value . $5 . $6 . $7;
					} else {#value not defined
						#only use the bare separators if there were framed ones as well
						$out  = (defined($2) || defined($3)) ? $1 : "";
						$out .= (defined($5) || defined($7)) ? $6 : "";
					}
					$out;
				}e;
	if ($str eq "" && defined($safestr)) {
		# if there isn't anything left of the format string after the replacements, use the safe string, if supplied
		return infoFormat($safestr,$file);
	} else {
		$str=~ s/%([0-9a-fA-F][0-9a-fA-F])%/chr(hex($1))/eg;
	}

	return $str;
}

#
# if no ID3 information is available,
# use this to get a title, which is derived from the file path or URL.
# Also used to get human readable titles for playlist files and directories.
#
# for files, file URLs and directories:
#             Any ending .mp3 is stripped off and only last part of the path
#             is returned
# for HTTP URLs:
#             URL unescaping is undone.
#

sub plainTitle {
	my ($file) = @_;

	my $title = "";

	$::d_info && SliMP3::Misc::msg("Plain title for: " . $file);

	if (isFileURL($file)) {
		$file = SliMP3::Misc::pathFromFileURL($file);
		if ($file) {
			$title = (splitdir($file))[-1];
		}
	} elsif (isHTTPURL($file)) {
		$title = SliMP3::HTTP::unescape($file);
	} else {
		$title = (splitdir($file))[-1];
	}
	
	if ($title) {
		$title =~ s/\.mp3$//io;
		$title =~ s/\.mp2$//io;
		$title =~ s/\.m3u$//io;
		$title =~ s/\.cue$//io;
		$title =~ s/\.pls$//io;
		$title =~ s/\.lnk$//io;
		$title =~ s/\.ogg$//io;
	}
	
	$::d_info && SliMP3::Misc::msg(" is " . $title . "\n");

	return $title;
}

# get a potentially client specifically formatted title.
sub standardTitle {
	my $client = shift;
	my $fullpath = shift;
	my $title;
	my $format;

	if (isITunesPlaylistURL($fullpath)) {
		$format = 'TITLE';
	} elsif (defined($client)) {
		#in array syntax this would be $titleFormat[$clientTitleFormat[$clientTitleFormatCurr]]
		#get the title format
		$format = SliMP3::Prefs::getInd("titleFormat"
				#at the array index of the client titleformat array
				,SliMP3::Prefs::clientGet($client, "titleFormat"
					#which is currently selected
					,SliMP3::Prefs::clientGet($client,'titleFormatCurr')));
	} else {
		#in array syntax this would be $titleFormat[$titleFormatWeb]
		$format = SliMP3::Prefs::getInd("titleFormat",SliMP3::Prefs::get("titleFormatWeb"));
	}
	
	$title = infoFormat($fullpath, $format, "TITLE");

	return $title;
}

#
# Return a structure containing the ID3 tag attributes of the given MP3 file.
#
sub infoHash {
	my $file = shift;

	if (!defined($file) || $file eq "") { 
		$::d_info && SliMP3::Misc::msg("trying to get infoHash on an empty file name\n");
		$::d_info && SliMP3::Misc::bt();
		return; 
	};
	my $item = cacheEntry($file);
	
	# we'll update the cache if we don't have a valid title in the cache
	if (!defined($item) && !defined(cacheItem($file, 'TITLE'))) {
		$::d_info && SliMP3::Misc::bt();
		$::d_info && SliMP3::Misc::msg("cache miss for $file\n");
		$item = updateCache($file)
	}
	
	return $item;

}

sub info {
	my $file = shift;
	my $tagname = shift;

	if (!defined($file) || $file eq "" || !defined($tagname)) { 
		$::d_info && SliMP3::Misc::msg("trying to get info on an empty file name\n");
		$::d_info && SliMP3::Misc::bt();
		return; 
	};
	
	my $item = cacheItem($file, $tagname);

	# we'll update the cache if we don't have a valid title in the cache
	if (!defined($item)) {
		# if it's the TITLE, then we need to do a full cache update, 
		#if we know the title but are missing some MP3::Info mp3 data, then look that up.  (Important for iTunes, as we don't get all the goods from the XML file)
		if (!defined(cacheItem($file, 'TITLE'))) {
			$::d_info && SliMP3::Misc::bt();
			$::d_info && SliMP3::Misc::msg("cache miss for $file\n");
			$item = updateCache($file)->{$tagname};
		} elsif ($tagname =~ /^(VERSION|LAYER|STEREO|VBR|BITRATE|FREQUENCY|SIZE|OFFSET|SECS|MM|SS|MS|TIME|COPYRIGHT|PADDING|MODE|FRAMES|FRAME_LENGTH|VBR_SCALE)$/) {
			updateMP3Info($file);
			$::d_info && SliMP3::Misc::msg("cache miss for MP3 info ($tagname) on $file\n");
			$item = cacheItem($file, $tagname);
		}		
	}	
	return $item;
}

sub trackNumber {
	my $file = shift;
	return (info($file,'TRACKNUM'));
}

sub cleanTrackNumber {
	my $tracknumber = shift;

	if (defined($tracknumber)) {
		#extracts the first digits only sequence then converts it to int
		$tracknumber =~ /(\d*)/;
		$tracknumber = $1 eq '' ? 0 : int($1);
	} else {
		$tracknumber = 0;
	}

	return $tracknumber;
}

sub genre {
	my $file = shift;
	return (info($file,'GENRE'));
}

sub title {
	my $file = shift;
	return (info($file,'TITLE'));
}

sub artist {
	my $file = shift;
	return (info($file,'ARTIST'));
}

sub composer {
	my $file = shift;
	return (info($file,'COMPOSER'));
}

sub album {
	my $file = shift;
	return (info($file,'ALBUM'));
}

sub year {
	my $file = shift;
	return (info($file,'YEAR'));
}

sub disc {
	my $file = shift;
	return (info($file,'DISC'));
}

sub discCount {
	my $file = shift;
	return (info($file,'DISCC'));
}

sub comment {
	my $file = shift;
	my $comment;
	my @comments;

	if (ref(info($file,'COMMENT')) eq 'ARRAY') {
		@comments = @{info($file,'COMMENT')};
	} else {
		@comments = (info($file,'COMMENT'));
	}
	# extract multiple comments and concatenate them
	if (@comments) {
		foreach my $c (@comments) {
			if ($c) {
				# ignore SoundJam CDDB comments
				if (
					$c =~ /SoundJam_CDDB_/ ||
					$c =~ /iTunes_CDDB_/ ||
					$c =~ /^\s*[0-9A-Fa-f]{8}(\+|\s)/
				) {
					#ignore
				} else {
					# put a slash between multiple comments.
					if ($comment) {
						$comment .= ' / ';
					}
					$c =~ s/^eng(.*)/$1/;
					$comment .= $c;
				}
			}
		}
	}
	return $comment;
}

sub duration {
	my $file = shift;
	my $secs = info($file,'SECS');

	if (defined $secs) {
		return sprintf('%s:%02s',int($secs / 60),$secs % 60);
	} else {
		return;
	}
}

sub durationSeconds {
	my $file = shift;
	return info($file,'SECS');
}

sub offset {
	my $file = shift;
	return info($file,'OFFSET');
}

sub size {
	my $file = shift;
	return info($file,'SIZE');
}

sub bitrate {
	my $file = shift;
	my $mode = (defined info($file,'VBR_SCALE')) ? 'VBR' : 'CBR';
	if (info($file,'BITRATE')) {
		return info($file,'BITRATE').SliMP3::Strings::string('KBPS').' '.$mode;
	} else {
		return;
	}
}

sub tagVersion {
	my $file = shift;
	return info($file,'TAGVERSION');
};

sub cachePlaylist {
	my $path = shift;

	my $inforef = infoHash($path);

	$inforef->{'LIST'} = shift;
	$inforef->{'AGE'} = Time::HiRes::time();
	updateCacheEntry($path, $inforef);
	
	$::d_info && SliMP3::Misc::msg("cached an " . (scalar @{$inforef->{'LIST'}}) . " item playlist for $path\n");
}

sub cachedPlaylist {
	my $path = shift;

	return info($path, 'LIST');
}

sub cachedPlaylistAge {
	my $path = shift;
	return info($path, 'AGE');
}

sub filterPrep {
	my $pattern = shift;
	#the following transformations assume that the pattern provided uses * to indicate
	#matching any character 0 or more times, and that . ^ and $ are not escaped
	$pattern =~ s/\\([^\*]|$)/($1 eq "\\")? "\\\\" : $1/eg; #remove single backslashes except those before a *
	$pattern =~ s/([\.\^\$\(\)\[\]\{\}\|\+\?])/\\$1/g; #escape metachars (other than * or \) in $pattern {}[]()^$.|+?
	$pattern =~ s/^(.*)$/\^$1\$/; #add beginning and end of string requirements
	$pattern =~ s/(?<=[^\\])\*/\.\*/g; #replace * (unescaped) with .*
	return qr/$pattern/i;
}

sub filterPats {
	my ($inpats) = @_;
	my @outpats = ();
	foreach my $pat (@$inpats) {
		push @outpats, filterPrep(uc($pat));
	}
	return \@outpats;
}

sub filter {
	my ($patterns, $const, @items) = @_;
	if (!defined($patterns) || ! @{$patterns}) {
		return @items;
	}

	my @filtereditems;
	# Gross, but this seems to be a relevant optimization.
	if ($const eq '') {
		ITEM: foreach my $item (@items) {
			foreach my $regexpattern (@{$patterns}) {
				if ($item !~ $regexpattern) {
					next ITEM;
				}
			}
			push @filtereditems, $item;
		}
	  } else {
		ITEM: foreach my $item (@items) {
			foreach my $regexpattern (@{$patterns}) {
				my $item_const = $item . ' ' . $const;
				if ($item_const !~ $regexpattern) {
					next ITEM;
				}
			}
			push @filtereditems, $item;
		}
	}

	return @filtereditems;
}

sub filterHashByValue {
	my ($patterns, $hashref) = @_;
	if (!defined($hashref)) {
		return;
	}

	if (!defined($patterns) || ! @{$patterns}) {
		return keys %{$hashref};
	}
	my @filtereditems;
	my ($k,$v);
	ENTRY: while (($k,$v) = each %{$hashref}) {
		foreach my $pat (@{$patterns}) {
			if ($v !~ $pat) {
				next ENTRY;
			}
		}
		push @filtereditems, $k;
	}

	return @filtereditems;
}

# genres|artists|albums|songs(genre,artist,album,song)
#===========================================================================
# Return list of matching keys at the given level in the genre tree.  Each
# of the arguments is an array reference of file glob type patterns to match.
# In order to match, all the elements of the list must match at the given
# level of the genre tree.

sub genres {
	my $genre = shift;
	my $artist = shift;
	my $album = shift;
	my $song = shift;

	my $genre_pats = filterPats($genre);
	return fixCase(sortuniq(filter($genre_pats,"",keys %genreCache)));
}

sub artists {
	my $genre = shift;
	my $artist = shift;
	my $album = shift;
	my $song = shift;
	my @artists = ();

	my $genre_pats = filterPats($genre);
	my $artist_pats = filterPats($artist);

	foreach my $g (filter($genre_pats,"",keys %genreCache)) {
		push @artists, filter($artist_pats,"",keys %{$genreCache{$g}});
	}

	return fixCase(sortuniq_ignore_articles(@artists));
}

sub albums {
	my $genre = shift;
	my $artist = shift;
	my $album = shift;
	my $song = shift;
	my @albums = ();

	my $genre_pats = filterPats($genre);
	my $artist_pats = filterPats($artist);
	my $album_pats = filterPats($album);

	foreach my $g (filter($genre_pats,"",keys %genreCache)) {
		foreach my $art (filter($artist_pats,"",keys %{$genreCache{$g}})) {
			if (SliMP3::Prefs::get("artistinalbumsearch")) {
				push @albums, filter($album_pats,$art,keys %{$genreCache{$g}{$art}});
			}
			else {
				push @albums, filter($album_pats,"",keys %{$genreCache{$g}{$art}});
			}
		}
	}

 	return fixCase(sortuniq_ignore_articles(@albums));
}

# return all songs for a given genre, artist, and album
sub songs {
	my $genre = shift;
	my $artist = shift;
	my $album = shift;
	my $track = shift;

	my $genre_pats = filterPats($genre);
	my $artist_pats = filterPats($artist);
	my $album_pats = filterPats($album);
	my $track_pats = filterPats($track);

	my @alltracks = ();

	$::d_info && SliMP3::Misc::msg("songs: $genre - $artist - $album - $track\n"	);

	foreach my $g (sortIgnoringCase(filter($genre_pats,'',keys %genreCache))) {
		foreach my $art (sortIgnoringCase(filter($artist_pats,'',keys %{$genreCache{$g}}))) {
			foreach my $alb (sortIgnoringCase(filter($album_pats,'',keys %{$genreCache{$g}{$art}}))) {
				my %songs = ();
				foreach my $trk (values %{$genreCache{$g}{$art}{$alb}}) {
					$songs{$trk} = title($trk);
				}
				push @alltracks, sort sortTrack filterHashByValue($track_pats,\%songs);
			}
		}
	}

	# remove duplicate tracks
	my %seen = ();
	my @uniq = ();
	
	foreach my $item (@alltracks) {
		push(@uniq, $item) unless (!defined($item) || ($item eq '') || $seen{uc($item)}++);
	}

	# if we are getting a specific album with an unspecific artist, re-sort the tracks
	if (scalar(@$album) == 1 && $album->[0] !~ /\*/  && (!defined($artist->[0]) || $artist->[0] eq '*')) {
		# if there are duplicate track numbers, then sort as multiple albums
		my $duptracknum = 0;
		my @seen = ();
		foreach my $item (@uniq) {
			my $trnum = trackNumber($item);
			if ($seen[$trnum]) {
				$duptracknum = 1;
				last;
			}
			$seen[$trnum]++;
		}
		if ($duptracknum) {
			@uniq =  sort sortTrack @uniq;
		} else {
			@uniq =  sort sortAlbum @uniq;
		}
	}

	return @uniq;
}

# sorting algorithm to use with the sort built-in function
# (This may be called from another package, so we use the parameter stack instead of $a and $b)
sub sortTrack ($$) {
	my $j = $_[0];
	my $k = $_[1];

	my $result;
	
	if (isList($j) && isList($k)) { return uc(title($j)) cmp uc(title($k)); };
	
	my $j_artist = artist($j);
	my $k_artist = artist($k);
	
	if (defined($j_artist) && defined($k_artist)) {
		$result = uc($j_artist) cmp uc($k_artist);
		if ($result) { return $result; }
	} elsif (defined($j_artist)) {
		return -1;
	} elsif (defined($k_artist)) {
		return 1;
	}
	
	my $j_album = album($j);
	my $k_album = album($k);
	
	if (defined($j_album) && defined($k_album)) {
		$result = uc($j_album) cmp uc($k_album);
		if ($result) { return $result; }
	} elsif (defined($j_album)) {
		return -1;
	} elsif (defined($k_album)) {
		return 1;
	}
	
	my $j_track = trackNumber($j);
	my $k_track = trackNumber($k);

	if ($j_track && $k_track) {
		$result = $j_track <=> $k_track;
		if ($result) { return $result; }
	} elsif ($j_track) {
		return -1;
	} elsif ($k_track) {
		return 1;
	}

	return uc(title($j)) cmp uc(title($k));
}

sub sortAlbum ($$) {
	my $j = $_[0];
	my $k = $_[1];

	my $result;
		
	if (isList($j) && isList($k)) { return uc(title($j)) cmp uc(title($k)); };
	
	my $j_album = album($j);
	my $k_album = album($k);
	
	if (defined($j_album) && defined($k_album)) {
		$result = uc($j_album) cmp uc($k_album);
		if ($result) { return $result; }
	} elsif (defined($j_album)) {
		return -1;
	} elsif (defined($k_album)) {
		return 1;
	}
	
	my $j_track = trackNumber($j);
	my $k_track = trackNumber($k);

	if ($j_track && $k_track) {
		$result = $j_track <=> $k_track;
		if ($result) { return $result; }
	} elsif ($j_track) {
		return -1;
	} elsif ($k_track) {
		return 1;
	}

	return uc(title($j)) cmp uc(title($k));
}

sub fileName {
	my $j = shift;

	if (isFileURL($j)) {
		$j = SliMP3::Misc::pathFromFileURL($j);
		if ($j) {
			$j = (splitdir($j))[-1];
		}
	} elsif (isHTTPURL($j)) {
		$j = SliMP3::HTTP::unescape($j);
	} else {
		$j = (splitdir($j))[-1];
	}
	return $j;
}

sub sortFilename ($$) {
	my $j = $_[0];
	my $k = $_[1];

	my $result;
	
	$j = fileName($j);
	$k = fileName($k);

	return uc($j) cmp uc($k);
}		


sub songPath {
	my $genre = shift;
	my $artist = shift;
	my $album = shift;
	my $track = shift;

	return $genreCache{uc($genre)}{uc($artist)}{uc($album)}{uc($track)};
}

sub isFragment {
	my $fullpath = shift;
	
	my $is = 0;
	if (isURL($fullpath)) {
		my $anchor = SliMP3::Misc::anchorFromURL($fullpath);
		if ($anchor && $anchor =~ /([\d\.]+)-([\d\.]+)/) {
			$is = 1;
		}
	}
}

sub updateCache {
	my $file = shift;
	my ($track, $song, $artistName,$albumName);
	my $filepath;
	my $type;
	my $tempCacheEntry;

	if (!defined($file) || $file eq "") { return; };

	# get the type without updating the cache
	$type = contentTypeFromPath($file);

	$::d_info && SliMP3::Misc::msg("Updating cache for: " . $file . "\n");

	# we only know how to extract ID3 information from file paths and file URLs.
	if ($type && $type=~/^mp[23]$/) {
		if ( (isFile($file) || isFileURL($file))) {
			if (isFileURL($file)) {
				$filepath = SliMP3::Misc::pathFromFileURL($file);
				
				# fragments, like in CUE files should just be ignored.  They have their own metadata.
				if (isFragment($file)) {
					$file =~ s/^([^#]*)#(.*)/$1/;
				}
			} else {
				$filepath = $file;
			}
			
			if ($filepath) {
				# get the MP3 tag information
				$tempCacheEntry = MP3::Info::get_mp3tag($filepath);
				$::d_info && !defined($tempCacheEntry) && SliMP3::Misc::msg("Info: no tags found for $file\n");
				# get the MP3 info information
				my $info = MP3::Info::get_mp3info($filepath);
				$::d_info && !defined($tempCacheEntry) && SliMP3::Misc::msg("Info: no mp3 info found for $file\n");
				
				# put everything we've got into $tempCacheEntry
				if ($info && $tempCacheEntry) {
					%{$tempCacheEntry} = (%{$tempCacheEntry}, %{$info});
				} elsif ($info && !$tempCacheEntry) {
					$tempCacheEntry = $info;
				}

				$tempCacheEntry->{'TRACKNUM'} = cleanTrackNumber($tempCacheEntry->{'TRACKNUM'});
				if ($tempCacheEntry->{'SET'}) {
					my $discNum = $tempCacheEntry->{'SET'};
					my $discCount;
					
					$discNum =~ /(\d+)\/(\d+)/;
					
					if ($1) {
						$discNum = $1;
					}
					
					$tempCacheEntry->{'DISCC'} = $discNum;
					
					if ($2) {
						$discCount = $2;
						$tempCacheEntry->{'DISCC'} = $discCount;
					}
					
					 my $discWord = string('DISC');
				    if ($discCount && ($discCount > 1) && ($tempCacheEntry->{'ALBUM'}!~ /(${discWord})|(Disc)\s+[0-9]+/i)) {
						# Add space to handle > 10 album sets and sorting. Is suppressed in the HTML.
						if ($discCount > 9 && $discNum < 10) { $discNum = ' ' . $discNum; };
							
						$tempCacheEntry->{'ALBUM'} = $tempCacheEntry->{'ALBUM'} . " ($discWord $discNum " . string('OF') . " $discCount)";
				    }
				}
				
				if (!$tempCacheEntry->{'TITLE'} && !defined(cacheItem($file, 'TITLE'))) {
					$::d_info && SliMP3::Misc::msg("Info: no title found, using plain title for $file\n");
					$tempCacheEntry->{'TITLE'} = plainTitle($file);					
				}

				# fix the genre
				if (defined($tempCacheEntry->{'GENRE'}) && $tempCacheEntry->{'GENRE'} =~ /^\((\d+)\)$/) {
					# some programs (SoundJam) put their genres in as text digits surrounded by parens.
					# in this case, look it up in the table and use the real value...
					if (defined($MP3::Info::mp3_genres[$1])) {
						$tempCacheEntry->{'GENRE'} = $MP3::Info::mp3_genres[$1];
					}
				}

				# cache the file size
				$tempCacheEntry->{'FS'} = -s $filepath;

				# cache the content type
				$tempCacheEntry->{'CT'} = $type;
				
				updateGenreCache($file,$tempCacheEntry);

			} else {
				$::d_info && SliMP3::Misc::msg("Info: no path found for $file\n");
				return;
			}
		} elsif (isHTTPURL($file)) {
			# if it's an HTTP URL, guess the title from the the last part of the URL,
			# and don't bother with the other parts
			if (!defined(cacheItem($file, 'TITLE'))) {
				$::d_info && SliMP3::Misc::msg("Info: no title found, calculating title from url for $file\n");
				if ($file =~ /([^\/]*)\.mp3\b/i) {
					$tempCacheEntry->{'TITLE'} = SliMP3::HTTP::unescape($1);
				} else {
					$tempCacheEntry->{'TITLE'} = SliMP3::HTTP::unescape($file)
				}
			}
			if (!defined(cacheItem($file, 'TRACKNUM'))) {
				$tempCacheEntry->{'TRACKNUM'} = 0;
			}
		}
	} elsif ($type && $type eq "ogg") {
		if ( (isFile($file) || isFileURL($file))) {
			if (isFileURL($file)) {
				$filepath = SliMP3::Misc::pathFromFileURL($file);
				
				# fragments, like in CUE files should just be ignored.  They have their own metadata.
				if (isFragment($file)) {
					$file =~ s/^([^#]*)#(.*)/$1/;
				}
			} else {
				$filepath = $file;
			}

			if ($filepath) {
				# get the Ogg comments
				$tempCacheEntry = SliMP3::Ogg::get_oggtag($filepath);
				$::d_info && !defined($tempCacheEntry) && SliMP3::Misc::msg("Info: no tags found for $file\n");
				
				$tempCacheEntry->{'TRACKNUM'} = cleanTrackNumber($tempCacheEntry->{'TRACKNUMBER'});
				if ($tempCacheEntry->{'SET'}) {
					my $discNum = $tempCacheEntry->{'SET'};
					my $discCount;
					
					$discNum =~ /(\d+)\/(\d+)/;
					
					if ($1) {
						$discNum = $1;
					}
					
					$tempCacheEntry->{'DISCC'} = $discNum;
					
					if ($2) {
						$discCount = $2;
						$tempCacheEntry->{'DISCC'} = $discCount;
					}
					
					 my $discWord = string('DISC');
				    if ($discCount && ($discCount > 1) && ($tempCacheEntry->{'ALBUM'}!~ /(${discWord})|(Disc)\s+[0-9]+/i)) {
						# Add space to handle > 10 album sets and sorting. Is suppressed in the HTML.
						if ($discCount > 9 && $discNum < 10) { $discNum = ' ' . $discNum; };
							
						$tempCacheEntry->{'ALBUM'} = $tempCacheEntry->{'ALBUM'} . " ($discWord $discNum " . string('OF') . " $discCount)";
				    }
				}
				
				if (!$tempCacheEntry->{'TITLE'} && !defined(cacheItem($file, 'TITLE'))) {
					$::d_info && SliMP3::Misc::msg("Info: no title found, using plain title for $file\n");
					$tempCacheEntry->{'TITLE'} = plainTitle($file);					
				}

				# cache the file size
				$tempCacheEntry->{'FS'} = -s $filepath;

				# cache the content type
				$tempCacheEntry->{'CT'} = $type;
								
				updateGenreCache($file,$tempCacheEntry);

			} else {
				$::d_info && SliMP3::Misc::msg("Info: no path found for $file\n");
				return;
			}
		} elsif (isHTTPURL($file)) {
			# if it's an HTTP URL, guess the title from the the last part of the URL,
			# and don't bother with the other parts
			if (!defined(cacheItem($file, 'TITLE'))) {
				$::d_info && SliMP3::Misc::msg("Info: no title found, calculating title from url for $file\n");
				if ($file =~ /([^\/]*)\.ogg\b/i) {
					$tempCacheEntry->{'TITLE'} = SliMP3::HTTP::unescape($1);
				} else {
					$tempCacheEntry->{'TITLE'} = SliMP3::HTTP::unescape($file)
				}
			}
			if (!defined(cacheItem($file, 'TRACKNUM'))) {
				$tempCacheEntry->{'TRACKNUM'} = 0;
			}
		}
        } else {
		my $title = plainTitle($file);
		$tempCacheEntry->{'TITLE'} = $title;
		$tempCacheEntry->{'TRACKNUM'} = 0;
	}
	
	if (!defined($tempCacheEntry->{'CT'})) {
		$tempCacheEntry->{'CT'} = $type;
	}
	
	updateCacheEntry($file, $tempCacheEntry);

	return $tempCacheEntry;
}

sub updateMP3Info {
	my $fullpath = shift;
	my $filepath;
	my $anchor;
	if (isFileURL($fullpath)) {
		$filepath = SliMP3::Misc::pathFromFileURL($fullpath);
		$anchor = SliMP3::Misc::anchorFromURL($fullpath);
	} else {
		$filepath = $fullpath;
	}
	if (-e $filepath && isMP3($fullpath)) {
		my $info = MP3::Info::get_mp3info($filepath);
		
		if ($anchor && $anchor =~ /([\d\.]+)-([\d\.]+)/) {
			my $start = $1;
			my $end = $2;
			
			my $duration = $end - $start;
			$::d_info && msg("updateMP3Info: calculating duration for anchor: $duration\n");
			my $size = $duration * SliMP3::Info::size($filepath) / SliMP3::Info::durationSeconds($filepath);
			$::d_info && msg("openSong: calculating sizae for anchor: $size\n");
			my $offset = SliMP3::Info::offset($filepath) + ($start * SliMP3::Info::size($filepath) / SliMP3::Info::durationSeconds($filepath));
			$::d_info && msg("openSong: calculating offset for anchor: $offset\n");
			$info->{'SIZE'} = $size;
			$info->{'OFFSET'} = $offset;
			$info->{'SECS'} = $duration;
		}

		updateCacheEntry($fullpath, $info);
	}
}

sub updateGenreCache {
	my $file = shift;
	my $cacheEntry = shift;

	# cache songs  uniquely
	my $genre = $cacheEntry->{'GENRE'};
	if (!defined ($genre) || !$genre) {
		$genre = 'Unknown';
	}

	my $artist = $cacheEntry->{'ARTIST'};
	if (!defined ($artist) || !$artist) {
		$artist = 'Unknown';
	}

	my $album = $cacheEntry->{'ALBUM'};
	if (!defined ($album) || !$album) {
		$album = 'Unknown';
	}

	my $track = cleanTrackNumber($cacheEntry->{'TRACKNUM'});

	if (!$track) {
		# we always have a title
		$track = $cacheEntry->{'TITLE'};
		if (!defined ($track) || !$track) {
			$track = 'Unknown';
		}
	}
	
	$genreCache{uc($genre)}{uc($artist)}{uc($album)}{uc($track)} = $file;

	$caseCache{uc($genre)} = $genre;
	$caseCache{uc($artist)} = $artist;
	$caseCache{uc($album)} = $album;
	$caseCache{uc($track)} = $track;
	
	my $composer = $cacheEntry->{'COMPOSER'};
	
	if ($composer) { 
		$genreCache{uc($genre)}{uc($composer)}{uc($album)}{uc($track)} = $file;
		$caseCache{uc($composer)} = $composer; 
	}
	
	$::d_info && SliMP3::Misc::msg("updating genre cache with: $genre - $artist - $album - $track\n--- for:$file\n");
}

sub addTitleToCache {
	my $url = shift;
	my $title = shift;

	$::d_info && SliMP3::Misc::msg("Adding title $title for $url\n");
	$::d_info && SliMP3::Misc::bt();
	
	my $cacheEntry = cacheEntry($url);

	$cacheEntry->{'TITLE'} = $title;
	if (!$cacheEntry->{'TRACKNUM'}) {
		$cacheEntry->{'TRACKNUM'} = 0;
	}
	updateCacheEntry($url, $cacheEntry);
}

sub fileLength {
	my $file = shift;
	return (info($file,'FS'));
}

sub isFile {
	my $fullpath = shift;

	$fullpath !~ /\.(?:mp2|mp3|m3u|pls|ogg|cue)$/i && return 0;

	my $stat = (-f $fullpath ? 1 : 0);

	$::d_info && SliMP3::Misc::msgf("isFile(%s) == %d\n", $fullpath, (1 * $stat));

	return $stat;
}

sub isFileURL {
	my $url = shift;

	return (defined($url) && ($url =~ /^file:\/\//i));
}

sub isITunesPlaylistURL { 
	my $url = shift;

	return (defined($url) && ($url =~ /^iTunesPlaylist:/i));
}

sub isHTTPURL {
	my $url = shift;

	return (defined($url) && ($url =~ /^http:\/\//i));
}

sub isURL {
	my $url = shift;

	return (defined($url) && (($url =~ /^file:\/\//i) || ($url =~ /^iTunesPlaylist:/i)  || ($url =~ /^http:\/\//i)));
}

sub isType {
	my $fullpath = shift;
	my $testtype = shift;
	my $type = contentType($fullpath);
	if ($type && ($type eq $testtype)) {
		return 1;
	} else {
		return 0;
	}
}

sub isWinShortcut {
	my $fullpath = shift;
	my $type = contentType($fullpath);
	return ($type && ($type eq 'lnk'));
}

sub isMP3 {
	my $fullpath = shift;
	my $type = contentType($fullpath);
	return ($type && (($type eq 'lnk') || ($type eq 'mp3') || ($type eq 'mp2')));
}

sub isOgg {
	my $fullpath = shift;
	my $type = contentType($fullpath);
	return ($type && ($type eq 'ogg'));
}

sub isSong {
	my $fullpath = shift;
	my $type = shift;
	if (!defined($type)) {
		$type = contentType($fullpath);
	}
	return ($type && (($type eq 'lnk') || ($type eq 'mp3') || ($type eq 'mp2') || ($type eq 'ogg')));
}

sub isDir {
	my $fullpath = shift;
	my $type = contentType($fullpath);
	return ($type && ($type eq 'dir'));
}

sub isM3U {
	my $fullpath = shift;
	my $type = contentType($fullpath);
	return ($type && ($type eq 'm3u'));
}

sub isPLS {
	my $fullpath = shift;
	my $type = contentType($fullpath);
	return ($type && ($type eq 'pls'));
}

sub isCUE {
	my $fullpath = shift;
	my $type = contentType($fullpath);
	return ($type && ($type eq 'cue'));
}

sub isList {
	my $fullpath = shift;
	my $is = 0;

	my $type = contentType($fullpath);
	return ($type && ( $type eq 'dir' || $type eq 'm3u' || $type eq 'pls' || $type eq 'cue' || $type eq 'lnk' || $type eq 'itu'));

# -- inlined!
#	if (isPlaylist($fullpath) ||
#		isDir($fullpath) ||
#		isWinShortcut($fullpath)
#	) {
#		$is = 1;
#	}
#
#	$::d_info && SliMP3::Misc::msgf("isList(%s) == %d\n", $fullpath, $is );
#
#	return $is;
}

sub isPlaylist {
	my $fullpath = shift;

	my $type = contentType($fullpath);
	return ($type && ( $type eq 'm3u' || $type eq 'pls' || $type eq 'cue' || $type eq 'itu'));

# -- inlined!
#	my $is = 0;
#	
#	if (isM3U($fullpath) ||
#		isPLS($fullpath) ||
#		isCUE($fullpath) ||
#		isITunesPlaylistURL($fullpath)) {
#		$is = 1;
#	}
#
#	$::d_info && SliMP3::Misc::msgf("isPlaylist(%s) == %d\n", $fullpath, $is );
#
#	return $is;
}

sub mimeType {
	my $file = shift;
	my $contentType = contentType($file);
	foreach my $mt (keys %mimeTypes) {
		if ($contentType eq $mimeTypes{	$mt }) {
			return $mt;
		}
	}
	return undef;
};


sub contentType {
	my $file = shift;
	return (info($file,'CT'));
}

sub contentTypeFromPath {
	my $fullpath = shift;
	my $type = 'unk';

	if (defined($fullpath) && $fullpath ne "" && $fullpath !~ /\x00/) {
		if (isHTTPURL($fullpath)) {
			if ($fullpath =~ /\.mp3$/i) {
				$type = 'mp3';
			} elsif ($fullpath =~ /\.mp2$/i) {
				$type = 'mp2';
			} elsif ($fullpath =~ /\.ogg$/i) {
				$type = 'ogg';
			} elsif ($fullpath =~ /\.m3u$/i) {
				$type = 'm3u';
			} elsif ($fullpath =~ /\.pls$/i) {
				$type = 'pls';
			} elsif ($fullpath =~ /\.cue$/i) {
				$type = 'cue';
			} else {
				# if we have no idea, then assume it's an MP3.  Once open the content-type will identify it for sure.
				$type = 'mp3';
			}
		} elsif (isITunesPlaylistURL($fullpath)) {
			$type = 'itu';
		} else {
			my $filepath;

			if (isFileURL($fullpath)) {
				$filepath = SliMP3::Misc::pathFromFileURL($fullpath);
				$::d_info && SliMP3::Misc::msg("Converting $fullpath to $filepath\n");
			} else {
				$filepath = $fullpath;
			}

#			$filepath = SliMP3::Misc::fixPath($filepath);
			if (defined($filepath) && $filepath ne "") {
				if (-f $filepath) {
					if ($filepath =~ /\.mp3$/i) {
						$type = 'mp3';
					} elsif ($filepath =~ /\.mp2$/i) {
						$type = 'mp2';
					} elsif ($filepath =~ /\.ogg$/i) {
						$type = 'ogg';
					} elsif ($filepath =~ /\.m3u$/i) {
						$type = 'm3u';
					} elsif ($filepath =~ /\.lnk$/i && SliMP3::OSDetect::OS() eq 'win') {
						require Win32::Shortcut;
						if ((Win32::Shortcut->new($filepath)) ? 1 : 0) {
							$type = 'lnk';
						}
					} elsif ($filepath =~ /\.pls$/i) {
						$type = 'pls';
					} elsif ($filepath =~ /\.cue$/i) {
						$type = 'cue';
					} else {
						$::d_info && SliMP3::Misc::msg("unknown file type for $filepath\n");
					}
				} elsif (-d $filepath) {
					$type = 'dir';
				} else {
				}
			}
		}
	}

	$::d_info && SliMP3::Misc::msg("$type file type for $fullpath\n");
	return $type;
}

sub sortIgnoringCase {
	return sort { uc($a) cmp uc($b) } @_;
}

sub fixCase {
	my @fixed = ();
	foreach my $item (@_) {
		push @fixed, $caseCache{$item};
	}
	return @fixed;
}

sub sortuniq {
	my %seen = ();
	my @uniq = ();

	foreach my $item (@_) {
		if (defined($item) && ($item ne '') && !$seen{uc($item)}++) {
			push(@uniq, $item);
		}
	}

	return sort { $a cmp $b } @uniq ;
}

# similar to above but ignore preceeding articles when sorting
sub sortuniq_ignore_articles {
	my %seen = ();
	my @uniq = ();
	my $articles =  SliMP3::Prefs::get("ignoredarticles");
	# allow a space seperated list in preferences (easier for humans to deal with)
	$articles =~ s/\s+/|/g;

	foreach my $item (@_) {
		if (defined($item) && ($item ne '') && !$seen{uc($item)}++) {
			push(@uniq, $item);
		}
	}

	return sort {
		my ($A,$B) = ($a,$b);
		$A =~ s/^($articles)\s+//i;
		$B =~ s/^($articles)\s+//i;
		return $A cmp $B;
	} @uniq ;
}

1;
__END__

# Local Variables:
# tab-width:4
# indent-tabs-mode:t
# End:
