package SliMP3::Strings;

# $Id: Strings.pm,v 1.20 2003/02/01 19:39:21 gr0tus Exp $

# SliMP3 Server Copyright (C) 2001 Sean Adams, Slim Devices Inc.
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License, 
# version 2.

use strict;
use vars qw($VERSION @ISA @EXPORT @EXPORT_OK @EXPORT_FAIL);
use File::Spec::Functions qw(:ALL);
require Exporter;
@ISA = qw(Exporter);
@EXPORT = qw( );
@EXPORT_OK = qw (string);    # we export string() so it's less typing to use it

#-------------------------------------------------

my %strings=();
my %languages=();
my $failsafe_language ="";

#
# Initializes the module
#
# strings_file is the file containing all the strings
#
# When a new string is added in strings.txt, it will probably take 
# a while before someome gets around to translating it. $failsafe_language
# is the fallback. If a string is not available in the user's
# preferred language (current_lang), then this is the one we'll return in it's place.
#

sub init {
	my ($strings_file, $failsafe) = @_;
	my $usr_strings;

	$SliMP3::Strings::failsafe_language = $failsafe;
	
	load_strings_file($strings_file); # First load the defaults
	
	my $userFile;
	if (SliMP3::OSDetect::OS() eq 'unix') {
		$userFile = ".slimp3-strings.txt";
	} else {
		$userFile = "slimp3-strings.txt"; 
	}
	
	$usr_strings = catdir(SliMP3::Prefs::preferencesPath(), $userFile);
	
	if ( -e $usr_strings ) { 
	    load_strings_file($usr_strings); # Then load any user preferences over the top
	}

    foreach my $lang (keys(%languages)) {
		$languages{$lang} = languageName($lang);
	}
}

#
# Loads a file containing strings
#
sub load_strings_file {

	my ($strings_file) = @_;
		
	open STRINGS, $strings_file || die "couldn't open $strings_file\n";
	
	my $string = '';
	my $language = '';
	my $stringname = '';
	my $line = '';
	my $ln = 0;

	LINE: while ($line=<STRINGS>) {
		$ln++;
		chomp($line);
		
		next if ($line=~/^#/);
		next if (!($line =~/\S/));

		if ($line=~/^(\S+)$/) {
			$stringname = $1;
			$string='';
			next LINE;
		} elsif ($line=~/^\t(\S*)\t(.+)$/) {
			my $one = $1;
			$string = $2;
			if ($one=~/./) {
				# if the string spans multiple lines, language can be left blank, and
				# we'll remember it from the last time we saw it.
				$language = $one;

				# keep track of all the languages we've seen
				# DISABLE RUSSION FOR NOW
				$languages{$language} = 1 if ($language ne 'RU');
				if (defined $strings{$language.'_'.$stringname}) { 
					delete $strings{$language.'_'.$stringname};
				};
			} 
			if (defined $strings{$language.'_'.$stringname}) {
				$strings{$language.'_'.$stringname} .= "\n".$string;
			} else {
				$strings{$language.'_'.$stringname} = $string;
			}
		} else {
			SliMP3::Misc::msg("Parse error on line $ln: $line\n");
		}
	}
	close STRINGS;
}


#
# Returns a list of all the available languages
#
sub list_of_languages {
	return sort(keys(%languages));
}

#
# Returns the hash of all available languages
#

sub hash_of_languages {
	return %languages;
}

#
# Returns a string in the requested language
#

sub string {
	my ($stringname) = @_;
	
	my $language = SliMP3::Prefs::get('language');

	if ($strings{$language.'_'.$stringname}) {
		return $strings{$language.'_'.$stringname};
	} else {
		if ($strings{$SliMP3::Strings::failsafe_language.'_'.$stringname}) {
			return $strings{$SliMP3::Strings::failsafe_language.'_'.$stringname};
		} else {
			warn "Undefined string: $stringname\nrequested language: $language\nfailsafe language: $SliMP3::Strings::failsafe_language\n";
			return '';
		}
	}
}

#
# Returns 1 if the requested string exists, 0 if not
#

sub stringExists {
	my $stringname = shift;
	if (!defined $stringname) {return 0;}
	my $language = SliMP3::Prefs::get('language');
	return ($strings{$language.'_'.$stringname} || $strings{$SliMP3::Strings::failsafe_language.'_'.$stringname}) ? 1 : 0;
}

#
# Sets the language in which strings will be returned
#
# returns 1 if the language is available, otherwise returns 0
# and current_lang is unchanged.
#
sub setLanguage {
	my $lang = shift;
	
	$lang=~tr/a-z/A-Z/;
	
	if (defined($languages{$lang})) {
		SliMP3::Prefs::set('language',$lang);
		return 1;
	}
	return 0;
}

sub getLanguage {
	return SliMP3::Prefs::get('language');
}

sub languageName {
	my $lang = shift;
	return $strings{$lang.'_LANGUAGE_CHOICES'};
}

#
# Returns the failsafe language:
#
sub failsafeLanguage {
	return $SliMP3::Strings::failsafe_language;
}

1;
